"""
    fulfill_from_stock!(
        x::SupplyChainEnv, src::Int, dst::Union{Int,Symbol}, 
        mat::Union{Symbol,String}, lead::Real
    )

Fulfill request from on-hand inventory.
"""
function fulfill_from_stock!(
    x::SupplyChainEnv, src::Int, dst::Union{Int,Symbol}, 
    mat::Union{Symbol,String}, lead::Int
)
    #check if partial fulfillment is allowed
    partial_fulfillment = check_fulfillment_type(x.network, src, dst, mat)
    
    #loop through orders on relevant arc
    orders_df = relevant_orders(x, src, dst, mat)
    for row in eachrow(orders_df)
        supply = x.tmp[src,mat,:on_hand] #available supply
        order_amount = row.amount
        accepted_inv = accepted_from_stock(order_amount, supply, partial_fulfillment)
        if accepted_inv > 0
            row.amount -= accepted_inv #update x.open_orders (deduct fulfilled part of the order)
            x.tmp[src,mat,:on_hand] -= accepted_inv #remove inventory from site
            push!(x.fulfillments, (row.id, x.period, row.arc, mat, accepted_inv, :sent)) #log order as sent
            if dst == :market
                push!(x.fulfillments, (row.id, x.period, row.arc, mat, accepted_inv, :delivered)) #log order as delivered
            else
                make_shipment!(x, row.id, src, dst, mat, accepted_inv, lead) #ship material (unless it is external demand)
            end
        end
        if x.options[:reallocate] #try to reallocate unfulfilled order if reallocation allowed
            reallocate_demand!(x, row)
        end
    end
    #remove any fulfilled orders from x.open_orders
    fulfilled_order_list = filter(:amount => <=(0), orders_df, view=true).id
    if !isempty(fulfilled_order_list)
        fulfilled_orders = Set(fulfilled_order_list) #find fulfilled orders
        if length(fulfilled_order_list) != length(fulfilled_orders) #there are production orders that were fulfilled from stock (delete these)
            prod_keys = Set([:consumption,:coproduction])
            filter!([:id,:arc] => (id,arc) -> !(id in fulfilled_orders && arc[2] in prod_keys), x.orders)
        end
        filter(:id => in(fulfilled_orders), x.orders, view=true).fulfilled .= x.period #store fulfillment date
        filter!(:id => !in(fulfilled_orders), x.open_orders) #remove fulfilled orders
    end
end

"""
    fulfill_from_production!(
        x::SupplyChainEnv, src::Int, dst::Int, mat::Union{Symbol,String}, 
        lead::Real, capacities::Dict
    )

Fulfill request by scheduling material production.
"""
function fulfill_from_production!(
    x::SupplyChainEnv, src::Int, dst::Union{Int,Symbol}, mat::Union{Symbol,String}, 
    lead::Int, capacities::Dict
)

    #check if partial fulfillment is allowed
    partial_fulfillment = check_fulfillment_type(x.network, src, dst, mat)

    #extract info
    bom = get_prop(x.network, src, :bill_of_materials)
    rmat_names = names(filter(<(0), bom[:,mat]), 1) #names of raw materials
    cmat_names = names(filter(>(0), bom[:,mat]), 1) #names of co-products

    #get relevant orders on that arc & raw material orders (:consumption)
    orders_df = relevant_orders(x, src, dst, mat) #NOTE: Assume that if plant requests production, it will accept early fulfillment
    raw_orders_grp = raw_material_orders(x, orders_df.id, rmat_names)
    #loop through orders on relevant arc
    for row in eachrow(orders_df)
        cap_and_sup = get_capacity_and_supply(x, src, mat, bom, rmat_names, cmat_names, capacities)
        order_amount = row.amount #amount requested in order
        accepted_prod = accepted_production(order_amount, cap_and_sup, partial_fulfillment) #amount accepted
        if accepted_prod > 0
            #fulfill order (may be partial)
            row.amount -= accepted_prod #update x.open_orders (deduct fulfilled quantity)
            #consume reactant
            for rmat in rmat_names
                consume_reactant!(x, row.id, src, rmat, bom[rmat,mat], accepted_prod, lead, raw_orders_grp) 
            end
            #schedule coproduction
            for cmat in cmat_names
                co_production!(x, row.id, src, dst, cmat, bom[cmat,mat], accepted_prod, lead, capacities)
            end
            #schedule production
            production!(x, row.id, src, dst, mat, accepted_prod, lead, capacities)
        end
        if x.options[:reallocate] #try to reallocate unfulfilled order if reallocation allowed
            reallocate_demand!(x, row)
        end
    end
    #remove any fulfilled orders from x.open_orders
    fulfilled_orders = Set(filter(:amount => <=(0), orders_df, view=true).id) #find fulfilled orders
    if !isempty(fulfilled_orders)
        filter(:id => in(fulfilled_orders), x.orders, view=true).fulfilled .= x.period #store fulfillment date
        filter!(:id => !in(fulfilled_orders), x.open_orders) #remove fulfilled orders (with any associated production orders)
    end
end

"""
    check_fulfillment_type(net::MetaDiGraph, src::Int, dst::Union{Int,Symbol}, mat::Union{Symbol,String})

Check if destination `dst` node accepts partial fulfillments.
"""
function check_fulfillment_type(net::MetaDiGraph, src::Int, dst::Union{Int,Symbol}, mat::Union{Symbol,String})
    #check if partial fulfillment is allowed
    src == dst && return true #NOTE: ASSUME if self-request (i.e., production order), then allow partial fulfillments
    indicator_node = dst == :market ? src : dst
    param_key = dst == :market ? :market_partial_fulfillment : :partial_fulfillment
    partial_fulfillment = get_prop(net, indicator_node, param_key)[mat]

    return partial_fulfillment
end

"""
    accepted_from_stock(order_amount, supply, partial_fulfillment)

Amount accepted from available supply on an order received.
"""
function accepted_from_stock(order_amount, supply, partial_fulfillment)
    if partial_fulfillment
        return min(order_amount, supply) #amount fulfilled from inventory
    else
        return order_amount <= supply ? order_amount : 0. #only accept full order or nothing at all
    end
end

"""
    accepted_production(order_amount, cap_and_sup, partial_fulfillment)

Amount of production order accepted based on available capacity, supply, and partial fulfillment criteria.
"""
function accepted_production(order_amount, cap_and_sup, partial_fulfillment)
    if partial_fulfillment
        return min(order_amount, cap_and_sup...) #fulfill remaining with available capacity
    else
        return order_amount <= minimum(cap_and_sup) ? order_amount : 0. #only accept full order or nothing at all
    end
end

"""
    consume_reactant!(x::SupplyChainEnv, order_id, src, rmat, stoich, accepted_prod, lead, raw_orders_grp)

Consume reactant `rmat` and update inventory and orders.
"""
function consume_reactant!(x::SupplyChainEnv, order_id, src, rmat, stoich, accepted_prod, lead, raw_orders_grp)
    consumed = accepted_prod * stoich #negative number
    x.tmp[src,rmat,:on_hand] += consumed
    raw_orders_grp[(id = order_id, material = rmat)].amount[1] += consumed
    push!(x.fulfillments, (order_id, x.period, (src,:consumption), rmat, -consumed, :sent)) #raw material is consumed ("sent")
end

"""
    co_production!(x::SupplyChainEnv, order_id, src, dst, cmat, stoich, accepted_prod, lead, capacities)

Produce coproduct `cmat` and update inventory and shipments
"""
function co_production!(x::SupplyChainEnv, order_id, src, dst, cmat, stoich, accepted_prod, lead, capacities)
    coproduction = accepted_prod*stoich
    capacities[cmat] -= coproduction #update coproduct production capacity
    push!(x.fulfillments, (order_id, x.period, (src,:coproduction), cmat, coproduction, :sent))
    make_shipment!(x, order_id, src, src, cmat, coproduction, lead) #schedule shipment of co-product 
end

"""
    production!(x::SupplyChainEnv, src, dst, mat, accepted_prod, lead, capacities)

Produce material `mat` and create shipments and update fulfillments/demand tables.
"""
function production!(x::SupplyChainEnv, order_id, src, dst, mat, accepted_prod, lead, capacities)
    capacities[mat] -= accepted_prod #update production capacity to account for commited capacity (handled first come first serve)
    push!(x.fulfillments, (order_id, x.period, (src,dst), mat, accepted_prod, :sent)) #update fulfillment log
    make_shipment!(x, order_id, src, dst, mat, accepted_prod, lead) 
end

"""
    make_shipment!(
        x::SupplyChainEnv, order_id::Int,
        src::Int, dst::Union{Int,Symbol}, mat::Union{Symbol,String}, 
        accepted_amount::Float64, lead::Float64
    )

Schedule shipment of material.
"""
function make_shipment!(
    x::SupplyChainEnv, order_id::Int,
    src::Int, dst::Union{Int,Symbol}, mat::Union{Symbol,String}, 
    accepted_amount::Float64, lead::Int
)
    #add inventory to pipeline
    a = (src,dst)
    if dst != :market #don't add if market order
        x.tmp[a,mat,:pipeline] += accepted_amount
    end
    push!(x.shipments, [order_id, a, mat, accepted_amount, lead]) #log the shipment
    iszero(lead) && update_shipments!(x) #if leadtime is zero, update shipments immediately
end

"""
    update_shipments!(x::SupplyChainEnv)

Update inventories throughout the network for arrived shipments.
"""
function update_shipments!(x::SupplyChainEnv)
    #find active shipments with 0 lead time
    arrivals = filter(:lead => <=(0), x.shipments, view=true) 
    for row in eachrow(arrivals)
        id, a, mat, amount = row[[:id, :arc, :material, :amount]]
        #update inventories and capacities
        shipment_completed!(x, id, amount, a, mat)
    end
    filter!(:lead => >(0), x.shipments) #remove shipments that arrived (and any zero values)
end

"""
    shipment_completed!(x::SupplyChainEnv, order_id::Int, amount::Float64, arc::Tuple, mat::Union{Symbol,String})

Update downstream inventory and pipeline inventory for completed shipment, and update production capacities for completed production.
"""
function shipment_completed!(x::SupplyChainEnv, order_id::Int, amount::Float64, arc::Tuple, mat::Union{Symbol,String})
    #update downstream inventory and pipeline inventory
    if arc[end] != :market
        x.tmp[arc[end],mat,:on_hand] += amount
        x.tmp[arc,mat,:pipeline] -= amount
    end
    push!(x.fulfillments, (order_id, x.period, arc, mat, amount, :delivered))
    #restore production capacities (production has concluded)
    if arc[1] == arc[2] || arc[2] == :market
        #restore mat production capacity
        capacity = get_prop(x.network, arc[1], :production_capacity)
        capacity[mat] += amount
        #restore any co-product production capacity
        bom = get_prop(x.network, arc[1], :bill_of_materials)
        cmat_names = names(filter(>(0), bom[:,mat]), 1) #names of co-products
        for cmat in cmat_names
            capacity[cmat] += amount*bom[cmat,mat]
        end
        #update production capacities
        set_prop!(x.network, arc[1], :production_capacity, capacity)
    end
end